@@ -740,7 +740,9 @@ static bool ValidateHeader(const MO3ContainerHeader &containerHeader)
 	{
 		return false;
 	}
-	if(containerHeader.musicSize <= sizeof(MO3FileHeader) || containerHeader.musicSize >= uint32_max / 2u)
+	// Due to the LZ algorithm's unbounded back window, we could reach gigantic sizes with just a few dozen bytes.
+	// 512 MB of music data (not samples) is chosen as a safeguard that is probably (hopefully) *way* beyond anything a real-world module will ever reach.
+	if(containerHeader.musicSize <= sizeof(MO3FileHeader) || containerHeader.musicSize >= 0x2000'0000)
 	{
 		return false;
 	}
@@ -1322,6 +1324,8 @@ bool CSoundFile::ReadMO3(FileReader &file, ModLoadingFlags loadFlags)
 				sample.uFlags.set(CHN_STEREO);
 
 			FileReader sampleData = file.ReadChunk(smpHeader.compressedSize);
+			if(!smpHeader.length)
+				continue;
 			const uint8 numChannels = sample.GetNumChannels();
 
 			if(compression == MO3Sample::smpDeltaCompression || compression == MO3Sample::smpDeltaPrediction)
@@ -1415,7 +1419,7 @@ bool CSoundFile::ReadMO3(FileReader &file, ModLoadingFlags loadFlags)
 			if(!sampleChunk.chunk.IsValid())
 				continue;
 
-			SAMPLEINDEX sharedOggHeader = smp + sampleChunk.sharedHeader;
+			SAMPLEINDEX sharedOggHeader = (smp + sampleChunk.sharedHeader > 0) ? static_cast<SAMPLEINDEX>(smp + sampleChunk.sharedHeader) : smp;
 			// Which chunk are we going to read the header from?
 			// Note: Every Ogg stream has a unique serial number.
 			// stb_vorbis (currently) ignores this serial number so we can just stitch
